﻿using System.ComponentModel;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace TestProject;

/// <summary>
/// 자식 프로세스 추적자
/// </summary>
public static class ChildProcessTracker
{
    //////////////////////////////////////////////////////////////////////////////////////////////////// Import
    ////////////////////////////////////////////////////////////////////////////////////////// Static
    //////////////////////////////////////////////////////////////////////////////// Private

    #region 작업 객체 생성하기 - CreateJobObject(jobAttributesHandle, name)

    /// <summary>
    /// 작업 객체 생성하기
    /// </summary>
    /// <param name="jobAttributesHandle">작업 특성 핸들</param>
    /// <param name="name">명칭</param>
    /// <returns>작업 핸들</returns>
    [DllImport("kernel32", CharSet = CharSet.Unicode)]
    private static extern IntPtr CreateJobObject(IntPtr jobAttributesHandle, string name);

    #endregion
    #region 작업 객체 정보 설정하기 - SetInformationJobObject(jobHandle, jobObjectInformationType, jobObjectInformationHandle, jobObjectInformationLength)

    /// <summary>
    /// 작업 객체 정보 설정하기
    /// </summary>
    /// <param name="jobHandle">작업 핸들</param>
    /// <param name="jobObjectInformationType">작업 객체 정보 타입</param>
    /// <param name="jobObjectInformationHandle">작업 객체 정보 핸들</param>
    /// <param name="jobObjectInformationLength">작업 객체 정보 길이</param>
    /// <returns>처리 결과</returns>
    [DllImport("kernel32")]
    private static extern bool SetInformationJobObject
    (
        IntPtr                   jobHandle,
        JobObjectInformationType jobObjectInformationType,
        IntPtr                   jobObjectInformationHandle,
        uint                     jobObjectInformationLength
    );

    #endregion
    #region 작업 객체에 프로세스 할당하기 - AssignProcessToJobObject(jobHandle, processHandle)

    /// <summary>
    /// 작업 객체에 프로세스 할당하기
    /// </summary>
    /// <param name="jobHandle">작업 핸들</param>
    /// <param name="processHandle">프로세스 핸들</param>
    /// <returns>처리 결과</returns>
    [DllImport("kernel32", SetLastError = true)]
    private static extern bool AssignProcessToJobObject(IntPtr jobHandle, IntPtr processHandle);

    #endregion

    //////////////////////////////////////////////////////////////////////////////////////////////////// Import
    ////////////////////////////////////////////////////////////////////////////////////////// Static
    //////////////////////////////////////////////////////////////////////////////// Private

    #region Field

    /// <summary>
    /// 작업 핸들
    /// </summary>
    private static readonly IntPtr _jobHandle;

    #endregion

    //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
    ////////////////////////////////////////////////////////////////////////////////////////// Static

    #region 생성자 - ChildProcessTracker()

    /// <summary>
    /// 생성자
    /// </summary>
    static ChildProcessTracker()
    {
        if(Environment.OSVersion.Version < new Version(6, 2))
        {
            return;
        }

        string jobName = $"ChildProcessTracker{Process.GetCurrentProcess().Id}";

        _jobHandle = CreateJobObject(IntPtr.Zero, jobName);

        JOB_OBJECT_BASIC_LIMIT_INFORMATION jobObjectBasicLimitInformation = new JOB_OBJECT_BASIC_LIMIT_INFORMATION();

        jobObjectBasicLimitInformation.LimitFlags = JobObjectLimit.JobObjectLimitKillOnJobClose;

        JOB_OBJECT_EXTENDED_LIMIT_INFORMATION jobObjectExtendedLimitInformation = new JOB_OBJECT_EXTENDED_LIMIT_INFORMATION();

        jobObjectExtendedLimitInformation.BasicLimitInformation = jobObjectBasicLimitInformation;

        int length = Marshal.SizeOf(typeof(JOB_OBJECT_EXTENDED_LIMIT_INFORMATION));

        IntPtr jobObjectInformationHandle = Marshal.AllocHGlobal(length);

        try
        {
            Marshal.StructureToPtr(jobObjectExtendedLimitInformation, jobObjectInformationHandle, false);

            if(!SetInformationJobObject(_jobHandle, JobObjectInformationType.ExtendedLimitInformation, jobObjectInformationHandle, (uint)length))
            {
                throw new Win32Exception();
            }
        }
        finally
        {
            Marshal.FreeHGlobal(jobObjectInformationHandle);
        }
    }

    #endregion

    //////////////////////////////////////////////////////////////////////////////////////////////////// Method
    ////////////////////////////////////////////////////////////////////////////////////////// Static
    //////////////////////////////////////////////////////////////////////////////// Public

    #region 프로세스 추가하기 - AddProcess(process)

    /// <summary>
    /// 프로세스 추가하기
    /// </summary>
    /// <param name="process">프로세스</param>
    public static void AddProcess(Process process)
    {
        if(_jobHandle != IntPtr.Zero)
        {
            bool success = AssignProcessToJobObject(_jobHandle, process.Handle);

            if(!success && !process.HasExited)
            {
                throw new Win32Exception();
            }
        }
    }

    #endregion
}